# Copyright 2008-2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://launchpad.net/filesync-server

"""The Storage network server testcase.

Base classes to do all the testing.
"""

import logging
import os
import sys
import time
import uuid

from functools import wraps
from StringIO import StringIO

from django.conf import settings
from OpenSSL import crypto
from twisted.internet import reactor, defer, ssl
from twisted.internet.protocol import connectionDone
from twisted.python.failure import Failure
from twisted.trial.unittest import TestCase as TwistedTestCase

from metrics import METER_UTILITY
from s4 import s4
from oops_datedir_repo import serializer_rfc822
from backends.filesync.data.testing.testcase import StorageDALTestCase
from backends.utils import create_test_user
from ubuntuone.storage.server.auth import DummyAuthProvider
from ubuntuone.storage.server.server import (
    PREFERRED_CAP,
    StorageServerService,
    configure_oops,
)
from ubuntuone.storage.server.logger import configure_logger, TRACE
from ubuntuone.storageprotocol import client
from ubuntuone.storageprotocol.client import (
    StorageClientFactory, StorageClient)
from ubuntuone.storageprotocol import request, protocol_pb2

from s3lib.contrib import http_client
http_client.in_testing = True

server_key = settings.API_SERVER_KEY
server_crt = settings.API_SERVER_CRT
server_crt_chain = settings.API_SERVER_CRT_CHAIN

log = logging.getLogger("fsyncsrvr.SyncDaemon.TEST")


def configure_logging(debug_tests=True):
    """Configure logging for the tests."""
    logger = logging.getLogger(settings.API_SERVER_LOGGER_NAME)
    if debug_tests:
        level = TRACE
    else:
        level = logging.INFO

    configure_logger(logger=logger, filename='filesync-server-tests.log',
                     level=level, propagate=False,
                     logger_class=logging.FileHandler)

# configure logging
configure_logging()


class FakeTimestampChecker(object):
    """Fake timestamp checker."""

    def get_faithful_time(self):
        """In the present we trust."""
        return int(time.time())

# need to patch this timestamp checker so it doesn't go to the real server
client.tx_timestamp_checker = FakeTimestampChecker()


class State(object):
    """An empty class to store state."""


class BaseProtocolTestCase(TwistedTestCase):
    """
    Reusable part of ProtocolTestCase
    """
    # do we need to create oops files?
    createOOPSFiles = False
    # should we use s4 with ssl?
    s4_use_ssl = False
    heartbeat_interval = 0
    timeout = 120

    @property
    def port(self):
        """The TCP port where the server listens."""
        return self.service.port

    @property
    def s4_port(self):
        """The port where S4 listens."""
        return self.s4_conn.getHost().port

    def get_oops(self):
        """get the oops generated by this test."""
        for date in os.listdir(settings.OOPS_PATH):
            for file in os.listdir(settings.OOPS_PATH + "/" + date):
                yield settings.OOPS_PATH + "/" + date + "/" + file

    @defer.inlineCallbacks
    def setUp(self):
        """Setup for testing."""
        # make sure we start with clean state
        yield super(BaseProtocolTestCase, self).setUp()
        log.info("starting test %s.%s", self.__class__.__name__,
                 self._testMethodName)
        self.s4_site = site = s4.server.Site(s4.Root())
        self.ssl_cert = crypto.load_certificate(
            crypto.FILETYPE_PEM, server_crt)
        if server_crt_chain:
            self.ssl_cert_chain = crypto.load_certificate(
                crypto.FILETYPE_PEM, server_crt_chain)
        else:
            self.ssl_cert_chain = None
        self.ssl_key = crypto.load_privatekey(crypto.FILETYPE_PEM, server_key)
        if self.s4_use_ssl:
            ssl_context_factory = s4.ContextFactory(
                self.ssl_cert, self.ssl_key)
            self.s4_conn = reactor.listenSSL(0, site, ssl_context_factory)
        else:
            self.s4_conn = reactor.listenTCP(0, site)

        if self.createOOPSFiles:
            location = self.mktemp()
            self.patch(settings, 'OOPS_PATH', location)
            oops_config = configure_oops()

            try:
                os.mkdir(location)
            except OSError:
                for path in self.get_oops():
                    os.unlink(path)
        else:
            # we don't need a real oops serializer, just dump all oops data to
            # stderr for the test run
            oops_config = configure_oops()

            def publish(report):
                report.setdefault("id", uuid.uuid4())
                serializer_rfc822.write(report, sys.stderr)

            oops_config.publishers = [publish]

        self._state = State()
        self.service = StorageServerService(
            0, "localhost", self.s4_port, self.s4_use_ssl,
            s4.AWS_DEFAULT_ACCESS_KEY_ID,
            s4.AWS_DEFAULT_SECRET_ACCESS_KEY,
            auth_provider_class=self.auth_provider_class,
            oops_config=oops_config, status_port=0,
            heartbeat_interval=self.heartbeat_interval)
        yield self.service.startService()

    @defer.inlineCallbacks
    def tearDown(self):
        """Tear down after testing."""
        yield self.service.stopService()
        yield self.s4_conn.stopListening()
        log.info("finished test %s.%s", self.__class__.__name__,
                 self._testMethodName)
        yield super(BaseProtocolTestCase, self).tearDown()


class ClientTestHelper(object):
    """helper for StorageClient used in tests."""

    def on_notification(self, share, node, hash):
        """Handle new notifications."""
        self.factory.pending_notifications -= 1
        if self.factory.pending_notifications < 0:
            self.test_fail(Exception("too many notifications"))
        else:
            if (self.factory.test_success and
                    self.factory.pending_notifications == 0):
                self._test_done("finished after notifications")

    def _test_done(self, result):
        """Real test done."""
        self.factory.cancel_timeout()
        self.transport.loseConnection()
        self.factory.test_deferred.callback(result)

    def kill(self):
        """Destroy this client without ending the test."""
        self.factory.cancel_timeout()
        self.transport.loseConnection()

    def test_done(self, result=None):
        """Finish test with success. Wait for pending notifications."""
        if (self.factory.pending_notifications == 0 and
                not self.factory.test_failed):
            self._test_done(result)
        self.factory.test_success = True

    def test_fail(self, result):
        """End test with error."""
        if not self.factory.test_failed:
            self.factory.test_failed = True
            if (self.factory.timeout.active() and
                    not self.factory.timeout.cancelled):
                self.factory.timeout.cancel()
            self.transport.loseConnection()
            if not self.factory.test_deferred.called:
                self.factory.test_deferred.errback(result)

    def check_doesnotexist(self, result):
        """Check for error."""
        error = result.check(request.StorageRequestError)
        if error is not None and result.getErrorMessage() == 'DOES_NOT_EXIST':
            self.test_done("ok")
        else:
            self.test_fail(result)

    # protocol API
    def connectionMade(self):
        """Setup and call callback."""
        self.connectionLostHandler = lambda _: None
        if self.factory.pending_notifications > 0:
            self.set_node_state_callback(self.on_notification)
        self.factory.callback(self)

    def connectionLost(self, reason=connectionDone):
        """Handle connection lost."""
        self.connectionLostHandler(reason)


class SimpleClient(StorageClient, ClientTestHelper):
    """Simple client that calls a callback on connection."""
    log = log

    def __init__(self, *args, **kwargs):
        """create the instance"""
        StorageClient.__init__(self, *args, **kwargs)
        self.messages = []

    def connectionMade(self):
        """Handle connection made"""
        StorageClient.connectionMade(self)
        ClientTestHelper.connectionMade(self)

    def connectionLost(self, reason=connectionDone):
        """Handle connection lost."""
        ClientTestHelper.connectionLost(self, reason=reason)

    def processMessage(self, message):
        self.messages.append(message)
        StorageClient.processMessage(self, message)


class SimpleFactory(StorageClientFactory):
    """A test oriented protocol factory."""

    protocol = SimpleClient


class FactoryHelper(object):
    """A StorageClientFactory wrapped with useful helper methods for testing
    different clients.
    """

    def __init__(self, cb_func, factory=SimpleFactory(), timeout=None,
                 wait_notifications=0, caps=None, **kwargs):
        """create the instance"""
        self.factory = factory
        self.cb_func = cb_func
        self.cb_kwargs = kwargs
        self.caps = caps
        if timeout is None:
            timeout = 120
        self.test_success = False
        self.test_failed = False
        self.pending_notifications = wait_notifications
        self.test_deferred = defer.Deferred()
        self.protocols = []
        self.timeout = reactor.callLater(timeout, self.error_shutdown)

    def cancel_timeout(self):
        """Cancel the timeout delayed call."""
        if self.timeout.active():
            self.timeout.cancel()

    @defer.inlineCallbacks
    def callback(self, client):
        """The callback to be executed when connected."""
        if self.caps is not None:
            request = yield client.set_caps(self.caps)
            assert request.accepted

        result = yield self.cb_func(client, **self.cb_kwargs)
        defer.returnValue(result)

    def buildProtocol(self, addr):
        """Create a protocol and keep a list of them."""
        p = self.factory.buildProtocol(addr)
        # replace the factory with self
        p.factory = self
        self.factory.client = p
        self.protocols.append(p)
        return p

    def error_shutdown(self):
        """Callback on timeout."""
        for p in self.protocols:
            p.transport.loseConnection()
        if self.test_success and self.pending_notifications > 0:
            self.test_deferred.errback(
                Exception("timeout waiting for notifications"))
        else:
            self.test_deferred.errback(Exception("timeout"))

    def __getattr__(self, name):
        """forward call to the factory"""
        return getattr(self.factory, name)


class ProtocolTestCase(BaseProtocolTestCase):
    """Setup the storage server on a random port.

    Keeps the port number on self.port so children classes can just write
    client code.

    """

    auth_provider_class = DummyAuthProvider
    factory_class = SimpleFactory

    def _save_state(self, key, value):
        """Store values to be accessed by deferred functions."""
        setattr(self._state, key, value)
        return value

    def callback_test(self, func, wait_notifications=0,
                      timeout=None, caps=PREFERRED_CAP,
                      add_default_callbacks=False, use_ssl=False, **kwargs):
        """Create a client and call callback on connection."""
        if add_default_callbacks:
            @wraps(func)
            def wrapped(client, **kwargs):
                """Wrapper which wires up test_done/test_fail."""
                d = func(client, **kwargs)
                d.addCallbacks(client.test_done, client.test_fail)
                return d
        else:
            wrapped = func
        f = FactoryHelper(wrapped, factory=self.buildFactory(),
                          wait_notifications=wait_notifications,
                          timeout=timeout, caps=caps, **kwargs)
        # there are 3 ways to connect to a server.
        # tcp and ssl will work in the tests
        if use_ssl:
            reactor.connectSSL("localhost", self.ssl_port, f,
                               ssl.ClientContextFactory())
        else:
            reactor.connectTCP("localhost", self.port, f)

        # https connect requires a working proxy and a server on
        # the default port running (we are not setting this up for
        # automated testing yet)
        # proxy_tunnel.connectHTTPS('localhost', 3128, "localhost", 20101, f,
        #     user="test", passwd="test")
        return f.test_deferred

    def assertFails(self, d, failure_name):
        """
        Fail unless a failure called failure_name (from
        protocol_pb2.Error) is thrown. If a different type of failure
        is thrown, it will not be caught, and the test case will be
        deemed to have suffered an error, exactly as for an unexpected
        failure.
        """
        if getattr(protocol_pb2.Error, failure_name, None) is None:
            raise ValueError("Unknown failure %r" % failure_name)

        def callback(*a):
            """Things worked. This is bad."""
            message = "expected %s failure, but nothing failed" % failure_name
            return Failure(AssertionError(message))

        def errback(failure):
            """Things broke. Lets see if this is good."""""
            if not failure.check(request.StorageProtocolError):
                return failure
            message = getattr(failure.value, 'error_message', None)
            if message is None:
                message = failure.value.message
            failure_type = message.error.type
            expected_type = getattr(protocol_pb2.Error, failure_name)
            if failure_type != expected_type:
                return failure

        d.addCallbacks(callback, errback)
        return d

    def buildFactory(self, *args, **kwargs):
        """build self.factory with the specified args and kwargs"""
        return self.factory_class(*args, **kwargs)


class TestWithDatabase(ProtocolTestCase, StorageDALTestCase):
    """A protocol test case that uses the database."""

    @defer.inlineCallbacks
    def setUp(self):
        """Setup."""
        yield super(TestWithDatabase, self).setUp()
        # most of these tests require a "test" bucket to be avaialble,
        # but don't create it using the s3 api...
        self.s4_site.resource._add_bucket("test")

        users = (
            (0, u'usr0', 'pass0'),
            (1, u'usr1', 'pass1'),
            (2, u'usr2', 'pass2'),
            (3, u'usr3', 'pass3'),
        )
        for user_id, username, password in users:
            user = create_test_user(
                id=user_id, username=username, password=password)
            setattr(self, username, user)
            # set the password in the object just as a test simplifier
            user.password = password

        # tune the config for this tests
        self.patch(settings, 'API_SERVER_MULTIPART_THRESHOLD', 1024 * 1024 * 5)
        self.patch(settings, 'API_SERVER_STORAGE_CHUNK_SIZE', 1024 * 64)

    @defer.inlineCallbacks
    def tearDown(self):
        """Tear Down."""
        METER_UTILITY.service_meter = None
        yield super(TestWithDatabase, self).tearDown()

    def save_req(self, req, name):
        """Save a request for later use."""
        setattr(self._state, name, req)
        return req


class BufferedConsumer(object):
    """Consumer that stores the content in a internal buffer."""

    def __init__(self, bytes_producer):
        """Create a BufferedConsumer."""
        self.producer = bytes_producer
        self.producer.consumer = self
        self.buffer = StringIO()

    def resumeProducing(self):
        """IPushProducer interface."""
        if self.producer:
            self.producer.resumeProducing()

    def stopProducing(self):
        """IPushProducer interface."""
        if self.producer:
            self.producer.stopProducing()

    def pauseProducing(self):
        """IPushProducer interface."""
        if self.producer:
            self.producer.pauseProducing()

    def write(self, content):
        """Part of IConsumer."""
        self.buffer.write(content)
